#include "StdAfx.h"
#include "GoogleMap.h"

#include <math.h>
#include <windows.h>
#include <wininet.h>

#define M_PI	(3.14159265)

int CGoogleMap::GetMercatorLongitude(double longi, int zoom)
{
	if(longi > 180.0) longi = longi - 360.0;
	if(longi < -180.0) longi = longi + 360.0;

	double res = (longi + 180.0) / 360.0;

	return (int)(res * pow(2.0, (double)zoom));
}

int CGoogleMap::GetMercatorLatitude(double lati, int zoom)
{
	double maxlat = M_PI;

	if(lati > 90) lati = lati - 180;
	if(lati < -90) lati = lati + 180;

	// degre -> radians
	double phi = M_PI * lati / 180.0;

	double res;
	//double temp = Math.Tan(Math.PI / 4 - phi / 2);
	//res = Math.Log(temp); 
	res = 0.5 * log((1+sin(phi))/(1-sin(phi)));
	double maxTileY = pow(2.0, (double)zoom);
	int result = (int)(((1-res/maxlat)/2)*(maxTileY));

	return (result);
}

string CGoogleMap::GetSatTileId(int row, int col, int zoom)
{
	string tileid = "t";
	double halflat = row; 
	double locxmin, locxmax, locymin, locymax, locxmoy, locymoy;

	locxmin = 0; 
	locxmax = pow(2.0, (double)zoom);

	locymin = 0; 
	locymax = pow(2.0, (double)zoom);

	for (int i = 0; i < zoom; i++)
	{
		locxmoy = (locxmax + locxmin) / 2.0;
		locymoy = (locymax + locymin) / 2.0;
		if ((halflat < locymin) || (halflat > locymax) ||
			(col < locxmin) || (col > locxmax)) {
				return ("transparent");
		}

		if (halflat < locymoy)
		{
			locymax = locymoy;
			if (col < locxmoy)
			{ /*q quadrant (top left)*/
				tileid += "q";
				locxmax = locxmoy;
			} else {/*r quadrant (top right)*/
				tileid += "r";
				locxmin = locxmoy;
			}
		} else {
			locymin = locymoy;
			if (col < locxmoy)
			{ /*t quadrant (bottom right)*/
				tileid += "t";
				locxmax = locxmoy;
			}
			else
			{/*s quadrant (bottom left)*/
				tileid += "s";
				locxmin = locxmoy;
			}
		}
	}

	return tileid;
}

string CGoogleMap::GetSatFullUrl(const string& _tileid, int srvnum)
{
	char szBuf[1024];
	//sprintf(szBuf, "http://kh%d.google.com/kh?n=404&v=15&t=%s" , srvnum, _tileid.c_str());
	sprintf(szBuf, "http://khm%d.google.com/kh?n=404&v=34&t=%s" , srvnum, _tileid.c_str());
	string fullurl(szBuf);

	if (_tileid == "transparent") {
		fullurl = "http://www.google.com/intl/en_ALL/mapfiles/transparent.png";
	}

	return fullurl;
}

void CGoogleMap::GetMinBounds(int numlat, int numlong, int zoom, double *pMinX, double *pMinY)
{
	double maxlat = M_PI;
	double maxTile = pow(2.0, (double)zoom);
	double numlatrel = maxlat * (1.0 - 2.0 * (((double)numlat + 1.0) / (double)maxTile));

	*pMinY = (double)((2.0 * atan(exp(numlatrel)) - M_PI / 2.0) * 180.0 / M_PI);
	*pMinX = (double)(360.0 * ((double)numlong / maxTile) - 180.0);
}

void CGoogleMap::GetMaxBounds(int numlat, int numlong, int zoom, double *pMaxX, double *pMaxY)
{
	double maxlat = M_PI;
	double maxTile = pow(2.0, (double)zoom);
	double numlatrel = maxlat * (1.0 - 2.0 * ((double)numlat / maxTile));

	*pMaxY = (double)((2.0 * atan(exp(numlatrel)) - M_PI / 2.0) * 180.0 / M_PI);
	*pMaxX = (double)(360.0 * ((double)numlong + 1) / maxTile - 180.0);
}


CGoogleMap::CGoogleMap(void)
{
}

CGoogleMap::~CGoogleMap(void)
{
}
