#include "StdAfx.h"
#include "gTrans.h"
#include <math.h>

void CgTrans::WGS84ToBesselTM(double Lati, double Long, double *pX, double *pY) 
{
	double bPhi, bLam, bh;

	WGP2BGP(Lati, Long, 0.0, &bPhi, &bLam, &bh);

	if(isKatech_) {
		BGP2KTM(bPhi, bLam, 0.0, pX, pY, 0);
	} else {
		BGP2KTM(bPhi, bLam, 10.405, pX, pY, 0);
		//BGP2KTM(bPhi, bLam, 0.0, pX, pY, 0);
	}
}

void CgTrans::BesselTMToWGS84(double X, double Y, double *pLati, double *pLong, int izone)
{
	double phi, lam, hW;

	if(isKatech_) {
		KTM2BGP(X, Y, 0.0, &phi, &lam, izone);
	} else {
		KTM2BGP(X, Y, 10.405, &phi, &lam, izone);
		//KTM2BGP(X, Y, 0.0, &phi, &lam, izone);
	}

	BGP2WGP(phi, lam, 0.0, pLati, pLong, &hW);

}

CgTrans::CgTrans(void) : AW_(6378137.0), FW_(1.0 / 298.257223563), AB_(6377397.155), FB_(1.0 / 299.1528153513233), OKUTM_(0.9996)
{
	SetTM();
}

void CgTrans::SetTM128() 
{
	falseNorthing_ = 600000.0;
	falseEasting_ =  400000.0;
	OKKTM_ = 0.9998;
	isKatech_ = true;
}

void CgTrans::SetTM() 
{
	falseNorthing_ = 500000.0;
	falseEasting_ =  200000.0;
	OKKTM_ = 1.0;
	isKatech_ = false;
}

CgTrans::~CgTrans(void)
{
}

void CgTrans::BGP2KTM(double phi, double lam, double dlam, double *x, double *y, int izone)
{
	// BESSEL  ѱ ǥ ȯ
	// Է
	//		phi, lam :   浵
	//		izone :  ȣ
	//				0 - ڵ 
	//				1 - ο (38, 125)
	//				2 - ߺο (38, 127)
	//				3 - ο (38, 129)
	//		dlam : 浵  , ʴ ( 10.405 = 0.0028902777 )
	//	
	//		x, y : ǥ (x : ,   y : )

	if(izone < 1) {
		izone = 2;
		if(lam < 126.0) izone = 1;
		if(lam > 128.0) izone = 3;
	}
	double centralMeridian;
	if(isKatech_)
		centralMeridian = 128.0 + dlam / 3600.0;
	else
		centralMeridian = 123.0 + (double)izone * 2.0 + dlam / 3600.0;

	GP2TM(phi, lam, AB_, FB_, falseNorthing_, falseEasting_, 38.0, centralMeridian, OKKTM_, x, y);
}

void CgTrans::GP2TM(double phi, double lam, double a, double f1, double x0, double y0, double phi0, double lam0, double ok, double *xn, double *ye)
{
/*
	' ----------------------------------------------------------------------------
	' ?FILE NAME: gp2utm.bas        VERSION: 9102.07      DESIGN DATE:  6/16/87 ?
	' ?STRUCTURE: procedure         PURPOSE: Compute UTM grid from GP           ?
	' ?LANGUAGE : Power Basic       AUTHOR : B. W. Drew                         ?
	' ѱ Ϲ TMǥȯ   ֵ ħ
	' zone ְ,  ǥ ô Է¹޾  99.12
	' ׷  561-3131
	'-----------------------------------------------------------------------------
	'
	' ?                           DOCUMENTATION                                 ?
	'-----------------------------------------------------------------------------
	' FUNCTIONAL DESCRIPTION:  Converts geographic (latitude, longitude)
	' coordinates to grid coordinates (zone, northing, easting) on the universal
	' transverse Mercator (UTM) projection. ACCURACY NOTE: Terms T4, T5, T8 & T9
	' may not be needed in applications requiring less accuracy.
	'
	' CALLING SEQUENCE:  call GP2UTM (A#, F#, SPHI#, SLAM#, IZONE, xn#, ye#, IFIXZ)
	' 99. 12     GP2TM(PHI#, LAM#, a#, f#, x0#, y0#, phi0#, lam0#, ok#, xn#, ye#)
	' INPUT ARGUEMENTS:
	'  Variable Name      Type  Description
	'  ------------------------------------
	'  phi#              dbl   latitude, radians - 99.12  
	'  lam#              dbl   longitude, radians 浵- 99.12  
	'  a#                 dbl   semi-major axis of ellipsoid, meters
	'  f#                 dbl   flattening of ellipsoid
	'  izone              int   UTM zone number - 99.12 
	'  ifixz              bool  Flag, = 0 compute UTM zone, = 1 use input zone - 99.12 
	' 99. 12 ߰
	' x0#, y0# :  ǥ
	' phi0#, lam0# :  , 浵
	' phi#, lam# : ȯϰ ϴ  , 浵
	' ok# :  ô
	'
	' OUTPUT ARGUEMENTS:
	'  Variable Name      Type  Description
	'  -------------------------------------
	'  izone              int   UTM zone number (see input arguements) - 99.12 
	'  xn#                 dbl   UTM northing, meters
	'  ye#                 dbl   UTM easting, meters
	'  99.12 )  x# , y# ̾ 츮   xn#  ħ
	'
	'
	'
	' FILES/DEVICES:  none
	'
	' ---------------------------------------------------------------------------
	' ?                              REVISIONS                                  ?
	' ---------------------------------------------------------------------------
	' REVISION   :  9102.07           AUTHOR:  R. E. Ziegler
	' DESCRIPTION:  Convert original FORTRAN to Power Basic.
	' 99. 12 visual basic  ħ, 
	' ---------------------------------------------------------------------------
	'
	'
	' 99.12 DIMEMSION
*/
	double sphi, slam, sphi0, slam0, f;
	double fe, degrad, recf, b, es, ebs, tn;
	double ap, bp, cp, dp, ep;
	double dLam, s, c, t, eta, sn, tmd, tmd1, nfn, xn1;
	double t1, t2, t3, t4, t5, t6, t7, t8, t9;

	fe = y0;     //  xǥ
	f = f1;       // 
	check(&f);       // 򵵰 299.1528...  Է Ǿ 0.003342.... ٲ

	degrad = atan(1.0) / 45.0;
	sphi = phi * degrad;
	slam = lam * degrad;
	sphi0 = phi0 * degrad;   //  
	slam0 = lam0 * degrad;

/*
	'    *************************************************
	'    *****   DERIVE OTHER ELLIPSOID PARAMTERS    *****
	'    *****         FROM INPUT VARIABLES          *****
	'    *****    A = SEMI-MAJOR AXIS OF ELLIPSOID   *****
	'    ***** RECF = RECIPROCAL OF FLATTENING (1/F) *****
	'    *************************************************
	'
	'
	'     ** SEMI MAJOR AXIS - B **
*/                 
	recf = 1.0 / f; //  299.1528...
	b = a * (recf - 1.0) / recf;

//	'
//	'     ** ECCENTRICITY SQUARED **

	es = (pow(a,2.0) - pow(b,2.0)) / pow(a,2.0);

//	'
//	'     ** SECOND ECCENTRICITY SQUARED **
	ebs = (pow(a,2.0) - pow(b,2.0)) / pow(b,2.0);
//	'
//	'     ** TRUE MERIDIONAL DISTANCE CONSTANTS **
	tn = (a - b) / (a + b);
	ap = a * (1.0 - tn + 5.0 * (pow(tn, 2.0) - pow(tn, 3.0)) / 4.0 + 81.0 * (pow(tn, 4.0) - pow(tn, 5.0)) / 64.0);
	bp = 3.0 * a * (tn - pow(tn, 2.0) + 7.0 * (pow(tn, 3.0) - pow(tn, 4.0)) / 8.0 + 55.0 * pow(tn, 5.0) / 64.0) / 2.0;
	cp = 15.0 * a * (pow(tn, 2.0) - pow(tn, 3.0) + 3.0 * (pow(tn, 4.0) - pow(tn, 5.0)) / 4.0) / 16.0;
	dp = 35.0 * a * (pow(tn, 3.0) - pow(tn, 4.0) + 11.0 * pow(tn, 5.0) / 16.0) / 48.0;
	ep = 315.0 * a * (pow(tn, 4.0) - pow(tn, 5.0)) / 512.0;

/*
	'
	'     ***** ZONE - CENTRAL MERIDIAN *****
	'
	'     *** ZONE ***
	'
	'     *** HOLD FIXED IF IFIXZ IS SET TO ONE
	' 99.12 Ʒ zone  
	'     *** DETERMINE ZONE NUMBER IF IFIXZ IS ZERO
	'    If IFIXZ = 0 Then
	'     ***
	'       IZONE = 31 + Int(SLAM# / DEGRAD# / 6#)
	'
	'         ** ZONE TRAP - AT HEMISPHERE LIMITS **
	'       If IZONE > 60 Then IZONE = 60
	'       If IZONE < 1 Then IZONE = 1
	'    End If
	'
	'     *** CENTRAL MERIDIAN ***
	'    OLAM# = (IZONE * 6 - 183) * DEGRAD#
	'
	'     *** DELTA LONGITUDE ***
	'     *** DIFFERENCE BETWEEN LONGITUDE AND CENTRAL MERIDIAN ***
*/
	dLam = slam - slam0;

/*
	'
	' ***   x ǥ nfn#      99.12
	'     ** TRUE MERIDIONAL DISTANCE **
*/
	tmd1 = fnSPHTMD(ap, bp, cp, dp, ep, sphi0);
	nfn = tmd1 * ok;

	//' ****  ϴ  x ǥ 
	//'     *** OTHER COMMON TERMS ***
	s = sin(sphi);
	c = cos(sphi);
	t = s / c;
	eta = ebs * pow(c, 2.0);

	// ** RADIUS OF CURVATURE IN PRIME VERTICAL **
	sn = fnSPHSN(a, es, sphi);

	// ** TRUE MERIDIONAL DISTANCE **
	tmd = fnSPHTMD(ap, bp, cp, dp, ep, sphi);

	//
	//     ***** NORTHING *****
	//
		
	t1 = tmd * ok;
	t2 = sn * s * c * ok / 2.0;
	t3 = sn * s * pow(c, 3.0) * ok * (5 - pow(t, 2.0) + 9 * eta + 4 * pow(eta, 2.0)) / 24;
	t4 = sn * s * pow(c, 5.0) * ok * (61 - 58 * pow(t, 2.0) + pow(t, 4.0) + 270 * eta - 330 * pow(t, 2.0) * eta 
		+ 445 * pow(eta, 2.0) + 324 * pow(eta, 3.0) - 680 * pow(t, 2.0) * pow(eta, 2.0) + 88 *pow(eta, 4.0)
		- 600 * pow(t, 2.0) * pow(eta, 3.0) - 192 * pow(t, 2.0) * pow(eta, 4.0)) / 720;
	t5 = sn * s * pow(c, 7) * ok * (1385.0 - 3111.0 * pow(t, 2.0) + 543 * pow(t, 4.0) - pow(t, 6)) / 40320.0;

	// ** FALSE NORTHING **

	xn1 = t1 + pow(dLam, 2.0) * t2 + pow(dLam, 4.0) * t3 + pow(dLam, 6.0) * t4 + pow(dLam, 8.0) * t5;
	*xn = xn1 - nfn + x0; //  ̸    99.12

	//     ***** EASTING *****
	t6 = sn * c * ok;
	t7 = sn * pow(c, 3.0) * ok * (1 - pow(t, 2.0) + eta) / 6.0;
	t8 = sn * pow(c, 5.0) * ok * (5.0 - 18.0 * pow(t, 2.0) + pow(t, 4.0) + 14.0 * eta - 58.0 * pow(t, 2.0) * eta
		+ 13.0 * pow(eta, 2.0) + 4.0 * pow(eta, 3.0) - 64.0 * pow(t, 2.0) * pow(eta, 2.0) - 24.0 * pow(t, 2.0) * pow(eta, 3.0)) / 120.0;
	t9 = sn * pow(c, 7.0) * ok * (61.0 - 479.0 * pow(t, 2.0) + 179.0 * pow(t, 4.0) - pow(t , 6.0)) / 5040.0;
	*ye = fe + dLam * t6 + pow(dLam, 3.0) * t7 + pow(dLam, 5.0) * t8 + pow(dLam, 7.0) * t9;
}

double CgTrans::fnSPHTMD(double ap, double bp, double cp, double dp, double ep, double sphi)
{
	// *** TRUE MERIDIONAL DISTANCE FROM LATITUDE ***

	return ap * sphi - bp * sin(2.0 * sphi) + cp * sin(4.0 * sphi) - dp * sin(6.0 * sphi) + ep * sin(8.0 * sphi);
}

void CgTrans::BGP2WGP(double phiB, double lamB, double hB, double *phiW, double *lamW, double *hW)
{
	// ѱ  WGS84  ȯ
	// Է
	//		phiB  : BESSEL 
	//		lamB  : BESSEL 浵
	//		hB    : BESSEL Ÿü 
	//		imode : ȯ
	//				1 : Molodensky 
	//				2 : Bursa 
	// 
	//		phiW  : WGS 84 
	//		lamW  : WGS 84 浵
	//		hW    : WGS 84 Ÿü 
		
	double XW, YW, zW, xb, yb, zB;

	//     ǥ ȯѴ
	BGP2BCTR(phiB, lamB, hB, &xb, &yb, &zB);

	//  ǥ WGS84 ǥ ȯѴ.
	BCTR2WCTR(xb, yb, zB, &XW, &YW, &zW);

	// WGS84 ǥ WGS84 ȯѴ
	WCTR2WGP(XW, YW, zW, phiW, lamW, hW);
}

void CgTrans::KTM2WGP(double xB, double yB, double hB, double dlam, double *phiW, double *lamW, double *hW, int izone)
{
	// ѱ ǥ WGS84  ȯ
	// Է
	//		phiB  : BESSEL 
	//		lamB  : BESSEL 浵
	//		hB    : BESSEL Ÿü 
	//		izone :  
	//				0: ڵ
	//				1: 
	//				2: ߺ
	//				3: 
	// 
	//		phiW  : WGS 84 
	//		lamW  : WGS 84 浵
	//		hW    : WGS 84 Ÿü 

	double phiB, lamB;

	//ѱ ǥ ѱ  ٲ
	KTM2BGP(xB, yB, dlam, &phiB, &lamB, izone);

	//ѱ  WGS84  ٲ
	BGP2WGP(phiB, lamB, hB, phiW, lamW, hW);
}

void CgTrans::KTM2BGP(double x, double y, double dlam, double *phi, double *lam, int izone)
{
	// ѱ ǥ BESSEL  ȯ
	// Է
	//		x, y : ǥ (x : ,   y : )
	//		izone :  ȣ
	//					0 - ڵ 
	//					1 - ο (38, 125)
	//					2 - ߺο (38, 127)
	//					3 - ο (38, 129)
	//		dlam : 浵  , ʴ ( 10.405 = 0.0028902777 )
	// 
	//		phi, lam :   浵

	if(izone<1) izone = 1;
	if(izone>3) izone = 3;

	double centralMeridian;
	if(isKatech_)
		centralMeridian = 128.0 + dlam / 3600.0;
	else
		centralMeridian = 123.0 + (double)izone * 2.0 + dlam / 3600.0;

	TM2GP(x, y, AB_, FB_, falseNorthing_, falseEasting_, 38.0, centralMeridian, OKKTM_, phi, lam);
}

double CgTrans::fnDENOM(double es, double sphi) 
{
	//     *** RADIUS OF CURVATURE IN THE MERIDIAN FROM LATITUDE ***

	return sqrt(1.0 - es * pow(sin(sphi), 2.0));
}

double CgTrans::fnSPHSR(double a, double es, double sphi)
{
	// *** RADIUS OF CURVATURE IN THE MERIDIAN FROM LATITUDE ***
	return a * (1.0 - es) / pow(fnDENOM(es, sphi), 3.0);
}

void CgTrans::TM2GP(double xn, double ye, double a, double f1, double x0, double y0, double phi0, double lam0, double ok, double *phi, double *lam)
{
/*
	' ----------------------------------------------------------------------------
	' ?FILE NAME: utm2gp.bas       VERSION : 9102.07      DESIGN DATE:  7/06/88 ?
	' ?STRUCTURE: Subroutine       AUTHOR  : B. W. Drew                         ?
	' ?LANGUAGE : Basic            COMPILER: PowerBASIC       VERSION: 2.10a    ?
	' ?PURPOSE  : Convert UTM to GP                                             ?
	' ѱ Ϲ TMǥȯ   ֵ ħ
	' zone ְ,  ǥ ô Է¹޾  99.12
	' ׷  561-3131
	' ----------------------------------------------------------------------------
	'
	' ----------------------------------------------------------------------------
	' ?                           DOCUMENTATION                                 ?
	' ----------------------------------------------------------------------------
	' FUNCTIONAL DESCRIPTION: CONVERTS GRID COORDINATES (ZONE, NORTHING, EASTING ) ***
	'  TO GEOGRAPHIC COORDINATES (LATITUDE, LONGITUDE) ON THE UNIVERSAL TRANSVERSE
	'  MERCATOR (UTM) GRID.
	'
	' CALLING SEQUENCE:  call UTM2GP(A#, F#, SPHI#, SLAM#, IZONE, xn#, ye#)
	'
	' INPUT ARGUEMENTS:
	'  Variable Name      Type  Description
	' ----------------------------------------------
	'  a#                 dbl   Ellipsoid semi-major axis, meters
	' f#                 dbl   Ellipsoid flattening
	' izone              int   UTM zone  -  99.12
	' xn#                 dbl   UTM northing, meters
	' ye#                 dbl   UTM easting, meters
	'
	' 99. 12 ߰
	' x0#, y0# :  ǥ
	' phi0#, lam0# :  , 浵
	' ok# :  ô

	' OUTPUT ARGUEMENTS:
	'  Variable Name      Type  Description
	'  ---------------------------------------------
	'  sphi#              dbl   Latitude, radians
	' slam#              dbl   Longitude, radians
	'
	' FILES/DEVICES: none
	' ----------------------------------------------------------------------------
	' ?                              REVISIONS                                  ?
	' ----------------------------------------------------------------------------
	' REVISION    : yymm.dd           AUTHOR: R. E. Ziegler
	' ORGANIZATION: x
	' DESCRIPTION : x
	' 99. 12 visual basic  ħ, 
	' ----------------------------------------------------------------------------
	'
	'99.12 DIMEMSION
*/
	double sphi, slam, sphi0, slam0, f;
	double fe, degrad, recf, b, es, ebs, tn;
	double ap, bp, cp, dp, ep;
	double dLam, s, c, t, eta, sn, tmd, tmd1, nfn, xn1;
	double t10, t11, t12, t13, t14, t15, t16, t17;
	double de, sr, ftphi;
	int i;

	f = f1;
	check(&f);

	fe = y0;
	degrad = atan(1.0) / 45.0;
	sphi0 = phi0 * degrad;
	slam0 = lam0 * degrad;

/*
	'    *************************************************
	'    *****   DERIVE OTHER ELLIPSOID PARAMTERS    *****
	'    *****         FROM INPUT VARIABLES          *****
	'    *****    A = SEMI-MAJOR AXIS OF ELLIPSOID   *****
	'    ***** RECF = RECIPROCAL OF FLATTENING (1/F) *****
	'    *************************************************
	'
	'     ** SEMI MAJOR AXIS - B **
*/
	recf = 1.0 / f;
	b = a * (recf - 1.0) / recf;

	// ** ECCENTRICITY SQUARED **
	es = (pow(a, 2.0) - pow(b, 2.0)) / pow(a, 2.0);

	// ** SECOND ECCENTRICITY SQUARED **
	ebs = (pow(a, 2.0) - pow(b, 2.0)) / pow(b, 2.0);

	// ** TRUE MERIDIONAL DISTANCE CONSTANTS **
	tn = (a - b) / (a + b);
	ap = a * (1.0 - tn + 5.0 * (pow(tn, 2.0) - pow(tn, 3.0)) / 4.0 + 81.0 * (pow(tn, 4.0) - pow(tn, 5.0)) / 64.0);
	bp = 3.0 * a * (tn - pow(tn, 2.0) + 7.0 * (pow(tn, 3.0) - pow(tn, 4.0)) / 8.0 + 55.0 * pow(tn, 5.0) / 64.0) / 2.0;
	cp = 15.0 * a * (pow(tn, 2.0) - pow(tn, 3.0) + 3.0 * (pow(tn, 4.0) - pow(tn, 5.0)) / 4.0) / 16.0;
	dp = 35.0 * a * (pow(tn, 3.0) - pow(tn, 4.0) + 11 * pow(tn, 5.0) / 16.0) / 48.0;
	ep = 315.0 * a * (pow(tn, 4.0) - pow(tn, 5.0)) / 512.0;
/*
	'     *** HEMISPHERE ADJUSTMENT TO FALSE NORTHING & POINT NORTHING ***
	'     *   NORTHERN HEMISPHERE
	'    nfn# = 0#
	'    If Sgn(xn#) < 0 Then
	'      nfn# = 10000000#
	'      xn# = Abs(xn#)
	'    End If
	'
	' ***   x ǥ nfn#      99.12
	'     ** TRUE MERIDIONAL DISTANCE **
*/
	tmd1 = fnSPHTMD(ap, bp, cp, dp, ep, sphi0);
	nfn = tmd1 * ok;

	// ** TRUE MERIDIONAL DISTANCE FOR FOOTPOINT LATITUDE **

	xn1 = xn + nfn - x0;
	tmd = xn1 / ok;

	//'     ***** FOOTPOINT LATITUDE *****
	//'
	//'     ** 1ST ESTIMATE **
	sr = fnSPHSR(a, es, 0.0);
	ftphi = tmd / sr;
/*	
		'     ******************************************
		'     ** ITERATE TO OBTAIN FOOTPOINT LATITUDE **
		'
*/
	for(i=1; i<=5; i++) {
		//'     * COMPUTED TRUE MERIDIONAL *
		t10 = fnSPHTMD(ap, bp, cp, dp, ep, ftphi);

		//'     * COMPUTED RADIUS OF CURVATURE IN THE MERIDIAN *
		sr = fnSPHSR(a, es, ftphi);

		//'     * CORRECTED FOOTPOINT LATITUDE *
		//'     * NEW FTPOINT = LAST FTPOINT +(TMDACTUAL -TMDCOMP)/RADIUS
		ftphi = ftphi + (tmd - t10) / sr;
	}

	//	'     ******************************************
	//	'
	//	'     ** RADIUS OF CURVATURE IN THE MERIDIAN **
	sr = fnSPHSR(a, es, ftphi);
	//	'
	//	'     ** RADIUS OF CURVATURE IN PRIME VERTICAL **
	sn = fnSPHSN(a, es, ftphi);
	//	'
	//	'     ** OTHER COMMON TERMS **
	s = sin(ftphi);
	c = cos(ftphi);
	t = s / c;
	eta = ebs * pow(c, 2.0);

	//	'     ** DELTA EASTING - DIFFERENCE IN EASTING **
	de = ye - fe;
	//	'     *******************************
	//	'
	//	'
	//	'     ***** LATITUDE *****
	//	'
	t10 = t / (2.0 * sr * sn * pow(ok, 2.0));
	t11 = t * (5.0 + 3.0 * pow(t, 2.0) + eta - 4.0 * pow(eta, 2.0) - 9.0 * pow(t, 2.0) * eta) / (24.0 * sr * pow(sn, 3.0) * pow(ok, 4.0));
	t12 = t * (61.0 + 90.0 * pow(t, 2.0) + 46.0 * eta + 45.0 * pow(t, 4.0) - 252.0 * pow(t, 2.0) * eta - 3.0 * pow(eta, 2.0) 
		+ 100.0 * pow(eta, 3.0) - 66.0 * pow(t, 2.0) * pow(eta, 2.0) - 90.0 * pow(t, 4.0) * eta + 88.0 * pow(eta, 4.0)
		+ 225.0 * pow(t, 4.0) * pow(eta, 2.0) + 84.0 * pow(t, 2.0) * pow(eta, 3.0) - 192.0 * pow(t, 2.0) * pow(eta, 4.0)) / (720.0 * sr * pow(sn, 5.0) * pow(ok, 6.0));
	t13 = t * (1385.0 + 3633.0 * pow(t, 2.0) + 4095.0 * pow(t, 4.0) + 1575.0 * pow(t, 6.0)) / (40320.0 * sr * pow(sn, 7.0) * pow(ok, 8.0));
		
	//     ** LATITUDE **
	sphi = ftphi - pow(de, 2.0) * t10 + pow(de, 4.0) * t11 - pow(de, 6.0) * t12 + pow(de, 8.0) * t13;

	//     ***** LONGITUDE *****
	t14 = 1.0 / (sn * c * ok);
	t15 = (1.0 + 2.0 * pow(t, 2.0) + eta) / (6.0 * pow(sn, 3.0) * c * pow(ok, 3.0));
	t16 = (5.0 + 6.0 * eta + 28.0 * pow(t, 2.0) - 3.0 * pow(eta, 2.0) + 8.0 * pow(t, 2.0) * eta + 24.0 * pow(t, 4.0) 
		- 4.0 * pow(eta, 3.0) + 4.0 * pow(t, 2.0) * pow(eta, 2.0) + 24.0 * pow(t, 2.0) * pow(eta, 3.0)) / (120.0 * pow(sn, 5.0) * c * pow(ok, 5.0));
	t17 = (61.0 + 662.0 * pow(t, 2.0) + 1320.0 * pow(t, 4.0) + 720.0 * pow(t, 6.0)) / (5040.0 * pow(sn, 7.0) * c * pow(ok, 7.0));

	//	'     ** DIFFERENCE IN LONGITUDE **
	dLam = de * t14 - pow(de, 3.0) * t15 + pow(de, 5.0) * t16 - pow(de, 7.0) * t17;

	//	'     ** CENTRAL MERIDIAN **
	//	'    olam# = (IZONE * 6 - 183) * DEGRAD#    ' 99.12
	//	'
	//	'     ** LONGITUDE **
	//	'    slam# = olam# + dlam#
	slam = slam0 + dLam;
	//	' change to degree unit

	*phi = sphi / degrad;
	*lam = slam / degrad;

	//	'     *** ACCURACY NOTE ***
	//	'     *** TERMS T12, T13, T16 & T17 MAY NOT BE NEEDED IN
	//	'     *** APPLICATIONS REQUIRING LESS ACCURACY
	//	'
}

void CgTrans::BCTR2WCTR(double xB, double yB, double zB, double *xW, double *yW, double *zW)
{
	// BESSEL ǥ WGS ǥ ٲ
	// Է
	//		xb, yb, zb :BESSEL ǥ
	// 
	//		xw, yw, zw :WGS ǥ

	if(TMODE_ == Bursa)
		InverseBursa(xB, yB, zB, xW, yW, zW);
	else
		InverseMolod(xB, yB, zB, xW, yW, zW);
}

void CgTrans::BGP2BCTR(double phi, double lam, double h, double *x, double *y, double *z)
{
	// BESSEL  BESSEL ǥ ٲ
	// Է
	//		phi# :  ()
	//		lam# : 浵
	//		h#   : Ÿü
	// 
	//		x#, y#, z# : ǥ

	GP2CTR(phi, lam, h, AB_, FB_, x, y, z);
}

void CgTrans::WGP2KTM(double phiW, double lamW, double hW, double dlam, double *xB, double *yB, double *hB, int izone)
{
	// WGS84  ѱ ǥ ȯ
	// Է
	//		phiW  : WGS 84 
	//		lamW  : WGS 84 浵
	//		hW    : WGS 84 Ÿü 
	//		imode : ȯ
	//				1 : Molodensky 
	//				2 : Bursa 
	//		izone :  
	//				0: ڵ
	//				1: 
	//				2: ߺ
	//				3: 
	// 
	//		phiB  : BESSEL 
	//		lamB  : BESSEL 浵
	//		hB    : BESSEL Ÿü 
		
	double phiB, lamB;

	// WGS84  ѱ  ٲ
	WGP2BGP(phiW, lamW, hW, &phiB, &lamB, hB);
	// ѱ  ѱ ǥ ٲ
	BGP2KTM(phiB, lamB, dlam, xB, yB, izone);
}

void CgTrans::SetPara(enumParameterMethod imode, double dx, double dy, double dz, double omega, double phi, double kappa, double ds)
{
	// ǥȯ 7 parameter 
	// dx,dy,dz :  
	// omega, phi, kappa : ȸ [ ]
	// ds : ô [1e-6 ]
	// XB = [1+ts]{R}XW + tx 
	// imode : ȯ
	//    1 : Molodensky 
	//    2 : Bursa 
	// Ʈ  ų  Ʈ  鶧
	// ǥȯ Ȯ 
	
	double degrad;

	degrad = atan(1.0) / 45.0;

	tx_ = dx;
	ty_ = dy;
	tz_ = dz;
	TOMEGA_ = omega / 3600.0 * degrad;
	tPhi_ = phi / 3600.0 * degrad;
	TKAPPA_ = kappa / 3600.0 * degrad;
	ts_ = ds * 0.000001;
	TMODE_ = imode;

	// set Bessel Geoid data
	xb_[1] = 34; yb_[1] = 126; hB_[1] = -50; wB_[1] = 3;
	xb_[2] = 35; yb_[2] = 126; hB_[2] = -55; wB_[2] = 2;
	xb_[3] = 36; yb_[3] = 126; hB_[3] = -60; wB_[3] = 2;
	xb_[4] = 37; yb_[4] = 126; hB_[4] = -64; wB_[4] = 2;
	xb_[5] = 38; yb_[5] = 126; hB_[5] = -65; wB_[5] = 3;
	xb_[6] = 34; yb_[6] = 127; hB_[6] = -46; wB_[6] = 2;
	xb_[7] = 35; yb_[7] = 127; hB_[7] = -49; wB_[7] = 1;
	xb_[8] = 36; yb_[8] = 127; hB_[8] = -56; wB_[8] = 1;
	xb_[9] = 37; yb_[9] = 127; hB_[9] = -63; wB_[9] = 1;
	xb_[10] = 38; yb_[10] = 127; hB_[10] = -65; wB_[10] = 2;
	xb_[11] = 34; yb_[11] = 128; hB_[11] = -43; wB_[11] = 2;
	xb_[12] = 35; yb_[12] = 128; hB_[12] = -45; wB_[12] = 1;
	xb_[13] = 36; yb_[13] = 128; hB_[13] = -52; wB_[13] = 1;
	xb_[14] = 37; yb_[14] = 128; hB_[14] = -58; wB_[14] = 1;
	xb_[15] = 38; yb_[15] = 128; hB_[15] = -62; wB_[15] = 2;
	xb_[16] = 34; yb_[16] = 129; hB_[16] = -42; wB_[16] = 2;
	xb_[17] = 35; yb_[17] = 129; hB_[17] = -42; wB_[17] = 1;
	xb_[18] = 36; yb_[18] = 129; hB_[18] = -47; wB_[18] = 1;
	xb_[19] = 37; yb_[19] = 129; hB_[19] = -52; wB_[19] = 1;
	xb_[20] = 38; yb_[20] = 129; hB_[20] = -60; wB_[20] = 2;
	xb_[21] = 34; yb_[21] = 130; hB_[21] = -42; wB_[21] = 3;
	xb_[22] = 35; yb_[22] = 130; hB_[22] = -42; wB_[22] = 2;
	xb_[23] = 36; yb_[23] = 130; hB_[23] = -44; wB_[23] = 2;
	xb_[24] = 37; yb_[24] = 130; hB_[24] = -48; wB_[24] = 2;
	xb_[25] = 38; yb_[25] = 130; hB_[25] = -53; wB_[25] = 3;

	// set WGS Geoid data
	XW_[1] = 34; YW_[1] = 126; hW_[1] = 24.5; wW_[1] = 3;
	XW_[2] = 35; YW_[2] = 126; hW_[2] = 23.8; wW_[2] = 2;
	XW_[3] = 36; YW_[3] = 126; hW_[3] = 23.3; wW_[3] = 2;
	XW_[4] = 37; YW_[4] = 126; hW_[4] = 22.7; wW_[4] = 2;
	XW_[5] = 38; YW_[5] = 126; hW_[5] = 22.9; wW_[5] = 3;
	XW_[6] = 34; YW_[6] = 127; hW_[6] = 26.2; wW_[6] = 2;
	XW_[7] = 35; YW_[7] = 127; hW_[7] = 26; wW_[7] = 1;
	XW_[8] = 36; YW_[8] = 127; hW_[8] = 25; wW_[8] = 1;
	XW_[9] = 37; YW_[9] = 127; hW_[9] = 24; wW_[9] = 1;
	XW_[10] = 38; YW_[10] = 127; hW_[10] = 23.8; wW_[10] = 2;
	XW_[11] = 34; YW_[11] = 128; hW_[11] = 27.8; wW_[11] = 2;
	XW_[12] = 35; YW_[12] = 128; hW_[12] = 28.5; wW_[12] = 1;
	XW_[13] = 36; YW_[13] = 128; hW_[13] = 28; wW_[13] = 1;
	XW_[14] = 37; YW_[14] = 128; hW_[14] = 27; wW_[14] = 1;
	XW_[15] = 38; YW_[15] = 128; hW_[15] = 25.8; wW_[15] = 2;
	XW_[16] = 34; YW_[16] = 129; hW_[16] = 28.8; wW_[16] = 2;
	XW_[17] = 35; YW_[17] = 129; hW_[17] = 29.3; wW_[17] = 1;
	XW_[18] = 36; YW_[18] = 129; hW_[18] = 29.7; wW_[18] = 1;
	XW_[19] = 37; YW_[19] = 129; hW_[19] = 28.8; wW_[19] = 1;
	XW_[20] = 38; YW_[20] = 129; hW_[20] = 27; wW_[20] = 2;
	XW_[21] = 34; YW_[21] = 130; hW_[21] = 29.3; wW_[21] = 3;
	XW_[22] = 35; YW_[22] = 130; hW_[22] = 29.4; wW_[22] = 2;
	XW_[23] = 36; YW_[23] = 130; hW_[23] = 29.4; wW_[23] = 2;
	XW_[24] = 37; YW_[24] = 130; hW_[24] = 28.9; wW_[24] = 2;
	XW_[25] = 38; YW_[25] = 130; hW_[25] = 27.8; wW_[25] = 3;
}

double CgTrans::Bgeoid(double phi, double lam)
{
	//  浵  Bessel geoid  
	// ̵   ̰ 
	//  Ű 󼭵 ̰ Ƿ Ű  ٲ
	//  function  Ǿ 
	//  function ġ̸,    ʿ

	int caltype;
	double gH = 0;
	int i;
	double L;
	double cl = 0;

	caltype = 2;

	switch(caltype) {
	case 1: //  1  Ǵ  p24, ȸ ׽ ̿, 
			// 츮  -42 ~ -65 ̳ -10 ~ -20 
		double u, v, k;
		k = 0.1570963;
		u = k * (phi - 35);
		v = k * (lam - 135);

		gH = 19.004 - 37.468 * u + 31.786 * v + 3.47 * u * v - 4.018 * pow(u, 3.0)
			+ 1.326 * u * pow(v, 2.0) - 8.48 * pow(v, 3.0) - 3.724 * pow(u, 3.0) * v + 7.243 * pow(u, 2.0) * pow(v, 2.0)
			- 4.923 * pow(v, 4.0) + 0.293 * pow(u, 9.0) * pow(v, 2.0) - 0.225 * pow(u, 5.0) * pow(v, 6.0)
			+ 0.2 * pow(u, 2.0) * pow(v, 9.0);
		
		break;
	case 2: // Bessel Geoid, ӿ  p70 ġȭ 
		for(i=1; i<=15; i++) {
			// ͸ ǥ ȯ
			L = sqrt( pow(phi - xb_[i], 2.0) + pow(lam - yb_[i], 2.0));
			if(L == 0) L = 0.00001;
			gH = gH + wB_[i] * hB_[i] / pow(L, 2.0);
			cl = cl + wB_[i] / pow(L, 2.0);
		}

		if(cl == 0) cl = 1; // error 
		gH = gH / cl;

		break;
	}
	
	return gH;
}

double CgTrans::Wgeoid(double phi, double lam)
{
	//  浵  WGS geoid  
	// ̵   ̰ 
	//  Ű 󼭵 ̰ Ƿ Ű  ٲ
	//  function  Ǿ 
	//  function ġ̸,    ʿ

	int caltype;
	double gH = 0.0;
	double X[10], y[10], h[10], L;
	double cl = 0.0;
	int i;

	caltype = 2;

	switch(caltype) {
	case 1: 
		// 2 Peter H. Dana, University of Texas at Austin  10 
		// geoid    Ǫ 
		// 40  120 gh=3,   130 gh= 23
		// 30  120 gh=7,   130 gh= 29   
		X[1] = 40; y[1] = 120; h[1] = 3;
		X[2] = 40; y[2] = 130; h[2] = 23;
		X[3] = 30; y[3] = 120; h[3] = 7;
		X[4] = 30; y[4] = 130; h[4] = 29;

		for(i=1; i<=4; i++) {
			// ͸ ǥ ȯ
			L = sqrt(pow((phi - X[i]), 2.0) + pow((lam - y[i]), 2.0));
			if(L == 0) L = 0.00001;
			gH = gH + h[i] / pow(L, 2.0);
			cl = cl + 1 / pow(L, 2.0);
		}

		gH = gH / cl;

		break;

	case 2:   // WGS Geoid, ӿ  p61 ġȭ 
		for(i=1; i<=15; i++) {
			// ͸ ǥ ȯ
			L = sqrt(pow(phi - XW_[i], 2.0) + pow((lam - YW_[i]), 2.0));
			if(L == 0) L = 0.00001;
			gH = gH + wW_[i] * hW_[i] / pow(L, 2.0);
			cl = cl + wW_[i] / pow(L, 2.0);
		}
		gH = gH / cl;
		break;
	}

	return gH;
}

void CgTrans::TransMolod(double xW, double yW, double zW, double *xB, double *yB, double *zB)
{
	// WGS ǥ  ǥ ȯ, Molodensky
	// Է
	//		xw,yw,zw : WGS ǥ
	// 
	//		xb,yb,zb :  ǥ
	//		XB = (1+ts){R}XW + tx 

	*xB = xW + (1 + ts_) * (TKAPPA_ * yW - tPhi_ * zW) + tx_;
	*yB = yW + (1 + ts_) * (-TKAPPA_ * xW + TOMEGA_ * zW) + ty_;
	*zB = zW + (1 + ts_) * (tPhi_ * xW - TOMEGA_ * yW) + tz_;
}

void CgTrans::TransBursa(double xW, double yW, double zW, double *xB, double *yB, double *zB)
{
	// WGS ǥ  ǥ ȯ, Bursa
	//		'Է
	//			'xw,yw,zw : WGS ǥ
	//		'
	//			'xb,yb,zb :  ǥ
	//			'XB = (1+ts){R}XW + tx 

	*xB = (1 + ts_) * (xW + TKAPPA_ * yW - tPhi_ * zW) + tx_;
	*yB = (1 + ts_) * (-TKAPPA_ * xW + yW + TOMEGA_ * zW) + ty_;
	*zB = (1 + ts_) * (tPhi_ * xW - TOMEGA_ * yW + zW) + tz_;
}

void CgTrans::WCTR2BCTR(double xW, double yW, double zW, double *xB, double *yB, double *zB)
{
	// WGS ǥ BESSEL ǥ ٲ
	// Է
	//		xw, yw, zw :WGS ǥ
	//		imode : ȯ
	//				1 : Molodensky 
	//				2 : Bursa 
	// 
	//		xb, yb, zb :BESSEL ǥ

	if(TMODE_ == Bursa) {
		TransBursa(xW, yW, zW, xB, yB, zB);
	} else {
		TransMolod(xW, yW, zW, xB, yB, zB);
	}
}

void CgTrans::WGP2BGP(double phiW, double lamW, double hW, double *phiB, double *lamB, double *hB)
{
	// WGS84  ѱ  ȯ
	// Է
	//		phiW  : WGS 84 
	//		lamW  : WGS 84 浵
	//		hW    : WGS 84 Ÿü 
	//		imode : ȯ
	//				1 : Molodensky 
	//				2 : Bursa 
	// 
	//		phiB  : BESSEL 
	//		lamB  : BESSEL 浵
	//		hB    : BESSEL Ÿü

	double XW, YW, zW, xb, yb, zB;

	//  WGS84 WGS84 ǥ ȯѴ
	WGP2WCTR(phiW, lamW, hW, &XW, &YW, &zW);

	// WGS84 ǥ  ǥ ȯѴ.
	WCTR2BCTR(XW, YW, zW, &xb, &yb, &zB);
	//  ǥ   ȯѴ
	BCTR2BGP(xb, yb, zB, phiB, lamB, hB);
}

void CgTrans::BCTR2BGP(double x, double y, double z, double *phi, double *lam, double *h)
{
	// BESSEL ǥ BESSEL  ٲ
	// Է
	//		x#, y#, z# : ǥ
	// 
	//		phi# :  ()
	//		lam# : 浵
	//		h#   : Ÿü
	
	CTR2GP(x, y, z, AB_, FB_, phi, lam, h);
}

void CgTrans::InverseBursa(double xB, double yB, double zB, double *xW, double *yW, double *zW)
{
	// ǥ  WGS ǥ  ȯ, Bursa
	// Է
	//		xb,yb,zb :  ǥ
	// 
	//		xw,yw,zw : WGS ǥ
	//		XW = 1/(1+ts) {R}^-1 {XB - tx} 

	double xt, yt, zt;

	xt = xB - tx_;
	yt = yB - ty_;
	zt = zB - tz_;

	*xW = 1.0 / (1.0 + ts_) * (xt - TKAPPA_ * yt + tPhi_ * zt);
	*yW = 1.0 / (1.0 + ts_) * (TKAPPA_ * xt + yt - TOMEGA_ * zt);
	*zW = 1.0 / (1.0 + ts_) * (-tPhi_ * xt + TOMEGA_ * yt + zt);
}

void CgTrans::InverseMolod(double xB, double yB, double zB, double *xW, double *yW, double *zW)
{
	// ǥ  WGS ǥ  ȯ, Molodensky
	// Է
	//		xb,yb,zb :  ǥ
	// 
	//		xw,yw,zw : WGS ǥ
	//	XW = 1/(1+ts) {R}^-1 {XB - tx} 

	double xt, yt, zt;

	xt = (xB - tx_) * (1.0 + ts_);
	yt = (yB - ty_) * (1 + ts_);
	zt = (zB - tz_) * (1 + ts_);

	*xW = 1.0 / (1.0 + ts_) * (xt - TKAPPA_ * yt + tPhi_ * zt);
	*yW = 1.0 / (1.0 + ts_) * (TKAPPA_ * xt + yt - TOMEGA_ * zt);
	*zW = 1.0 / (1.0 + ts_) * (-tPhi_ * xt + TOMEGA_ * yt + zt);
}

void CgTrans::GP2WGP(double phiU, double lamU, double hU, double a, double f, double *phiW, double *lamW, double *hW)
{
	//  Ÿü   WGS84  ȯ
	// Է
	//		phiU  :  Ÿü 
	//		lamU  :  Ÿü 浵
	//		hU    :  Ÿü 
	//		imode : ȯ
	//				1 : Molodensky 
	//				2 : Bursa 
	// 
	//		phiW  : WGS 84 
	//		lamW  : WGS 84 浵
	//		hW    : WGS 84 Ÿü 
	
	double XW, YW, zW, xU, yU, zU;

	//   Ÿü   Ÿü ǥ ȯѴ
	GP2CTR(phiU, lamU, hU, a, f, &xU, &yU, &zU);

	//  Ÿü ǥ WGS84 ǥ ȯѴ.
	if(TMODE_ == Bursa) {
		InverseBursa(xU, yU, zU, &XW, &YW, &zW);
	} else {
		InverseMolod(xU, yU, zU, &XW, &YW, &zW);
	}
	
	// WGS84ǥ WGS84  ȯѴ
	WCTR2WGP(XW, YW, zW, phiW, lamW, hW);
}

void CgTrans::WCTR2WGP(double x, double y, double z, double *phi, double *lam, double *h)
{
	// WGS ǥ WGS  ٲ
	// Է
	//		x#, y#, z# : ǥ
	// 
	//		phi# :  ()
	//		lam# : 浵
	//		h#   : Ÿü

	CTR2GP(x, y, z, AW_, FW_, phi, lam, h);
}

void CgTrans::WGP2GP(double phiW, double lamW, double hW, double a, double f, double *phiU, double *lamU, double *hU)
{
	// WGS84    Ÿü  ȯ
	// Է
	//		phiW  : WGS 84 
	//		lamW  : WGS 84 浵
	//		hW    : WGS 84 Ÿü 
	//		imode : ȯ
	//				1 : Molodensky 
	//			    2 : Bursa 

	// 
	// phiU  :  Ÿü 
	// lamU  :  Ÿü 浵
	// hU    :  Ÿü 
	double XW, YW, zW, xU, yU, zU;

	// WGS84 WGS84 ǥ ȯѴ
	WGP2WCTR(phiW, lamW, hW, &XW, &YW, &zW);

	//WGS84 ǥ  Ÿü ǥ ȯѴ.
	if(TMODE_ == Bursa) {
		TransBursa(XW, YW, zW, &xU, &yU, &zU);
	} else {
		TransMolod(XW, YW, zW, &xU, &yU, &zU);
	}
	
	//  Ÿü ǥ  Ÿü  ȯѴ
	CTR2GP(xU, yU, zU, a, f, phiU, lamU, hU);
}

void CgTrans::CTR2GP(double x, double y, double z, double a, double f1, double *phi, double *lam, double *h)
{
	// ǥ  ǥ ٲ
	// Է parameter
	//		x#, y#, z# : ǥ  x(north), y(east), z
	//		a#    : ݰ
	//		f#    :  (1/299.....)
	//  
	//		phi#  : 
	//		lam#  : 浵

	double degrad, sphiold, sphinew, slam, recf, b, es, n, p, t1, f;
	int icount;

	f = f1;
	check(&f);
   
	degrad = atan(1.0) / 45.0;
   
	// ܹݰ b#, Ⱦ ݰ N 
	//     ** SEMI MAJOR AXIS - B **
	recf = 1.0 / f;
	b = a * (recf - 1.0) / recf;
	es = (pow(a,2.0) - pow(b,2.0)) / pow(a,2.0);

	// 浵 
	slam = atan(y/x);
   
	//   Ÿü  
	p = sqrt(x * x + y * y);
   
	// 1 
	n = a;
	*h = 0;
	sphiold = 0;

	// 10^-12̳ ϶ ݺ
	for(icount=0; icount<30; icount++) {
		t1 = pow(( pow(b,2.0) / pow(a,2.0) * n + (*h)), 2.0) - pow(z,2.0);
		t1 = z / sqrt(t1);
		sphinew = atan(t1);
		if(abs(sphinew - sphiold) < (1e-18))
			break;
		//  Ⱦ ݰ  h
		n = fnSPHSN(a, es, sphinew);
		*h = p / cos(sphinew) - n;
		sphiold = sphinew;
	}

	*phi = sphinew / degrad;
	*lam = slam / degrad;
	if(x < 0.0) *lam = 180.0 + *lam; // 90 - 270 
	if(*lam < 0.0) *lam = 360.0 + *lam; // 270 - 360 
}

void CgTrans::WGP2WCTR(double phi, double lam, double h, double *x, double *y, double *z)
{
	// WGS  WGS ǥ ٲ
	// Է
	//		phi# :  ()
	//		lam# : 浵
	//		h#   : Ÿü
	// 
	//		x#, y#, z# : ǥ

	GP2CTR(phi, lam, h, AW_, FW_, x, y, z);
}

void CgTrans::check(double *f)
{
	// 򵵰 299.1528...  Է Ǿ 0.003342.... ٲ
	if(*f > 1.0) *f = 1 / (*f);
}

double CgTrans::fnSPHSN(double a, double es, double sphi)
{
// *** RADIUS OF CURVATURE IN THE PRIME VERTICAL FROM LATITUDE ***
	return a / sqrt(1.0 - es * pow(sin(sphi),2));
}

void CgTrans::GP2CTR(double phi, double lam, double h, double a, double f1, double *x, double *y, double *z)
{
	//  ǥ ٲ
	// Է parameter
	//		phi#  : 
	//		lam#  : 浵
	//		a#    : ݰ
	//		f#    :  (1/299.....)
	//  
	//		x#, y#, z# : ǥ  x(north), y(east), z
		
	double degrad, sphi, slam, recf, b, es, n, f;

	f = f1;
	check(&f);
	degrad = atan(1.0) / 45.0;
	sphi = phi * degrad;
	slam = lam * degrad;

	// ܹݰ b#, Ⱦ ݰ N 
	// ** SEMI MAJOR AXIS - B **
	recf = 1.0 / f;
	b = a * (recf - 1) / recf;
	es = (pow(a,2.0) - pow(b,2.0)) / pow(a,2.0);

	// Ⱦ ݰ
	n = fnSPHSN(a, es, sphi);

	// x, y, z 
	*x = (n + h) * cos(sphi) * cos(slam);
	*y = (n + h) * cos(sphi) * sin(slam);
	*z = ((pow(b,2.0) / pow(a,2.0)) * n + h) * sin(sphi);
}